
// union.cc
// Time-stamp: </wrf/c/union/union.cc, Sun, 1 Dec 2002, 22:22:56 EST,
// wrf@benvolio.ecse.rpi.edu>

// Debugging:
// #define DPC
// #define DV

// Print a status every 10 seconds if WATCHDOG is defined
#define WATCHDOG

#include "header"

// The universe is a square from (1,1) to (32767,32767).

const short int MAXCOORD = 32767;

short int edgelen;		// Length of all the edges, as an int, not a
				// fraction.
int     nsquares;		// Number of squares.
int     nedges;			// Number of edges, 4*nsquares.
short int gridsize;		// Number of grid cells on a side.
double  gridscale;		// Inverse of cell width.
int     ncovers;		// number of times a cell was covered.
int     nuniquecovers;		// number of covered cells.
int     nedgesegs;		// Number of (edge, cell) pairs, calling a 'segment' 
				// 
				// 
				// an occurrance of an edge in a cell.
int     ncoverededgesegs;	// Number of segments in a completely covered cell.
int     nreallocs;		// Number of times a cell's edge array had to be
				// reallocated to make it bigger. 
int     nreallocedgecopies;	// Number of edges copied during those
				// reallocations.
int     nxsect;			// Number of unique edge-edge intersections found.
				// N.b. Intersections in covered cells not found.
int     nxsectincdup;		// Number of intersections, including duplicates
				// caused by the same pair of intersecting edges
				// passing thru more than one cell.
int     nvisxsect;		// Number of visible intersections.
int     cx0, cy0, cx1, cy1;	// The ll and ur cells of the current square.
long int area;			// Computed area of the union of all the squares.
long int perimeter;		// Computed perimeter of the union of all the
				// squares.

struct timeval *watchdog_tv;
struct timezone *watchdog_tz;

struct Square
{
    short int x, y;		// lower left corner
};

Square *squares;

// All edges are the same length, edgelen.  Generate 4 edges from each
// square, in this order: 0-bottom, 1-left, 2-top, 3-right.  Edges are
// referred to by a number from 0 to nedges-1.  Don't bother to orient
// the edges so that inside is on the left since we never need that.

inline const short int
edge_x0 (const int i)
{
    return squares[i / 4].x + (i % 4 == 3 ? edgelen : 0);
}
inline const short int
edge_y0 (const int i)
{
    return squares[i / 4].y + (i % 4 == 2 ? edgelen : 0);
}
inline const short int
edge_x1 (const int i)
{
    return squares[i / 4].x + (i % 4 == 1 ? 0 : edgelen);
}
inline const short int
edge_y1 (const int i)
{
    return squares[i / 4].y + (i % 4 == 0 ? 0 : edgelen);
}


class   Cell
{
    int    *edges;		// The edges in this cell.

  public:

    short int m;		// number of edges allocated for this cell.
    short int nn;		// number of edges in this cell. -1 means this cell
    // is covered by a square.

            Cell ()		// Create an empty cell.
    {
	nn = 0;
	m = 0;
	edges = 0;
    }

    int    &operator[] (const int i) const	// Return the i-th edge in this
	// cell.
    {
	if (i < 0 || i >= nn)
	    die("cell[i]: i out of range");
	return edges[i];
    }

    void    cover ()		// Mark this cell as completely covered by some
	// square.
    {
	// cout << PRINTC(nn) << PRINTC(m) << PRINTN(edges);
	ncovers++;
	if (nn >= 0)
	    nuniquecovers++;
	if (m > 0)
	  {
	      ncoverededgesegs += nn;
	      delete[]edges;
	      edges = 0;
	  }
	nn = -1;
	m = 0;
    }

    void    append (const int e)
    {				// append a new edge to the cell.
	nedgesegs++;
	if (nn < 0)
	  {
	      ncoverededgesegs++;
	      return;		// cell is covered by a square.
	  }
	if (nn >= m)
	  {
	      int    *edges2;
	      if (m > 20000)
		{
		  die("Error: more than 20000 edges segments in a cell; make the grid finer and rerun.");
		}
	      m = static_cast < short int >(m * 1.5) + 3;
	      // cout << "Reallocating cell to hold " << m << " edges." << endl;
	      nreallocs++;
	      NEWA (edges2, int, m);
	      if (m > 0)
		{
		    nreallocedgecopies += nn;
		    for (int i = 0; i < nn; i++)
			edges2[i] = edges[i];
		    delete[]edges;
		}
	      edges = edges2;
	  }
	edges[nn++] = e;
    }


    int     n () const		// Return the number of edges in this cell.
    {
	return nn;
    }
};

Cell  **grid;


// PTIME Write cumulative CPU process time.

void
ptime (const char *const msg)
{
    float   t = ((float) clock ()) / CLOCKS_PER_SEC;
    cout << "Cumulative CPU time thru " << msg << "= " << t << endl;
}


void perr(char *msg) {   
  if (errno!=0) 
    {   perror(msg); exit(2);}
}


#ifdef WATCHDOG
// WATCHDOG_PRINT Maybe print, to cerr, a message, followed by endl.
// Print if at least PERIOD=10 seconds have elapsed since the last time
// that watchdog_print printed a message.

void watchdog_print(const char* const msg) {
  const int period = 10;
  static long lastprinttime=0;
  int ret;

  ret=gettimeofday(watchdog_tv,watchdog_tz);
  if (ret!=0) perr("gettimeofday called from watchdog_print");
  if (watchdog_tv->tv_sec-lastprinttime > period) {
    cerr << msg << endl;
    lastprinttime = watchdog_tv->tv_sec;
  }
}
#endif


void
initialize (const int argc, const char **const argv)
{
    print_env (cout);

    if (argc < 4)
      {
	  cerr << "argc=" << argc << endl;
	  die ("CONNECT requires at 3 arguments: nsquares, edgelen, gridsize.");
      }

    nsquares = atoi (argv[1]);
    edgelen = atoi (argv[2]);
    gridsize = atoi (argv[3]);

    nedges = nsquares * 4;
    gridscale = static_cast < double >(gridsize) / static_cast < double >(MAXCOORD);

    // initialize counters here instead of in the declarations so that
    // they will be reinitialized if I restart the program in the
    // debugger.

    ncovers = 0;
    nuniquecovers = 0;
    nedgesegs = 0;
    ncoverededgesegs = 0;
    nreallocs = 0;
    nreallocedgecopies = 0;
    nxsect = 0;
    nxsectincdup = 0;
    nvisxsect = 0;

    PRINTIOCC (nedges);
    PRINTIOCC (gridsize);
    PRINTIOCC (MAXCOORD);
    PRINTIOCC (edgelen);
    cout << PRINTN (gridscale) << endl;

    if (nsquares < 1 || nsquares > 2000000000)
	die ("illegal nsquares.");
    if (edgelen < 1 || edgelen > 10000)
	die ("Illegal edgelen.");
    if (gridsize < 1 || gridsize > 20000)
	die ("Illegal gridsize.");
    if (edgelen * gridscale < 1)
	die ("Grid cells must be smaller than edges.");

    NEWA (squares, Square, nsquares);

    area = 0;
    perimeter = 0;

#ifdef WATCHDOG
  watchdog_tv = new struct timeval;
  watchdog_tz = new struct timezone;
#endif
}

void
setup_grid ()
{
    NEWA (grid, Cell *, gridsize);
    for (int i = 0; i < gridsize; i++)
      {
	  NEWA (grid[i], Cell, gridsize);
      }
}

void
make_random_squares ()
{
    for (int i = 0; i < nsquares; i++)
      {
	  double  r;
	  Square & square = squares[i];

	  // random returns in [0,RAND_MAX].  I want the range to be
	  // [1,MAXCOORD-edgelen].

	  r = static_cast < double >(random ());
	  r = r *
	      static_cast < double >(MAXCOORD - edgelen - 1) / static_cast <
	      double >(RAND_MAX) + 1;
	  square.x = max (1, min (MAXCOORD - edgelen, static_cast < int >(r)));

	  r = static_cast < double >(random ());
	  r = r *
	      static_cast < double >(MAXCOORD - edgelen - 1) / static_cast <
	      double >(RAND_MAX) + 1;
	  square.y = max (1, min (MAXCOORD - edgelen, static_cast < int >(r)));

	  // cout << "Making square: (" << square.x << ',' << square.y << ')' <<
	  // endl;
      }
}


template < class C > inline short int
determine_1d_cell (C xy)
{
    return min (static_cast < short int >(xy * gridscale),
		static_cast < short int >(gridsize - 1));
}


void
put_edge_into_cells (const int e, const int lcx0, const int lcy0,
		     const int lcx1, const int lcy1)
{
    // Pass in the endpoint cells, instead of calculating here, for speed.

    // cout << "Putting edge " << e << " into cells:";

    // This could be done faster, if necessary.

    for (int ix = lcx0; ix <= lcx1; ix++)
	for (int iy = lcy0; iy <= lcy1; iy++)
	  {
	      // cout << " (" << ix << ',' << iy << ") ";
	      grid[ix][iy].append (e);
	  }
    // PRINTIOC(nedgesegs);
    // cout << endl;
}


void
put_square_and_edges_into_cells (const int i)
{
    Square & square = squares[i];
    const int x0 = square.x;
    const int y0 = square.y;
    const int x1 = square.x + edgelen;
    const int y1 = square.y + edgelen;

#ifdef WATCHDOG
    if (i%100000==0) {
    ostrstream sout;
    sout << "put_square_and_edge_into_cells, i=" << i << ends;
    watchdog_print(sout.str());
    }
#endif

    cx0 = determine_1d_cell (x0);	// cx0 etc are global to simplify
    // communication.
    cy0 = determine_1d_cell (y0);
    cx1 = determine_1d_cell (x1);
    cy1 = determine_1d_cell (y1);

    // cout << "Square's cell coords: (" << cx0 << ',' << cy0 << ") to (" << cx1 <<
    // ',' << cy1 << ')' << endl;

    for (int ix = cx0 + 1; ix < cx1; ix++)
	for (int iy = cy0 + 1; iy < cy1; iy++)
	  {
	      // cout << "Square covers cell (" << ix << ',' << iy << ") " << endl;
	      grid[ix][iy].cover ();
	  }

    put_edge_into_cells (4 * i, cx0, cy0, cx1, cy0);
    put_edge_into_cells (4 * i + 1, cx0, cy0, cx0, cy1);
    put_edge_into_cells (4 * i + 2, cx0, cy1, cx1, cy1);
    put_edge_into_cells (4 * i + 3, cx1, cy0, cx1, cy1);
}



// Test if (x,y) is visible, that is, is not contained in any square.
// To resolve degeneracies, we also need the numbers of the squares
// that x and y came from.  Also, a point cannot be hidden by a square
// whose edge created this point.

bool
visible (const long int x, const long int y, const int xsquare, const int ysquare)
{
    const short int cx = determine_1d_cell (x);
    const short int cy = determine_1d_cell (y);
    const   Cell & cell = grid[cx][cy];
    const int n (cell.n ());

#ifdef DV
    cout << "Testing visibility ";
    PRINTIOCC (x);
    PRINTIOCC (y);
    PRINTIOCC (xsquare);
    PRINTIOCC (ysquare);
    PRINTIOC (n);
    cout << endl;
#endif

    if (n < 0)			// Cell is covered; everything in it is hidden.
	return false;
    if (n == 0)
	die ("Testing a point's visibility, but no edges are in this cell. How can this be?");

    // Check if (x,y) is on the inside of an edge segment in this cell.
    // (NOT on ths inside of the infinite line containing the edge).
    // This test works only because the cell size is smaller than the
    // edge length.

    for (int i = 0; i < n; i++)
      {
	  const int &edge = cell[i];
	  const int dir = edge % 4;	// b, l, t, r
	  const int ex0 = edge_x0 (edge);
	  const int ey0 = edge_y0 (edge);
	  const int ex1 = edge_x1 (edge);
	  const int ey1 = edge_y1 (edge);
	  const int esquare = edge / 4;

#ifdef DV
	  PRINTIOCC (i);
	  PRINTIOCC (edge);
	  PRINTIOC (esquare);
	  cout << endl;
#endif
	  if (esquare == xsquare || esquare == ysquare)
	      continue;

#define LESS(a,b,ea,eb) ( (a)<(b) || (a)==(b) && (ea)<(eb) )

#ifdef DV
#define RET { cout << "(" << x << "," << y << ") hidden by edge " << edge << endl; return false; }
#else
#define RET return false;
#endif

	  switch (dir)
	    {
	    case 0:		// bottom edge.  Is it below point?
		if (LESS (ey0, y, esquare, ysquare)
		    && LESS (ex0, x, esquare, xsquare)
		    && LESS (x, ex1, xsquare, esquare))
		    RET     break;
	    case 1:		// left (vertical) edge.  
		if (LESS (ex0, x, esquare, xsquare)
		    && LESS (ey0, y, esquare, ysquare)
		    && LESS (y, ey1, ysquare, esquare))
		    RET     break;
	    case 2:		// top edge.  Is it above point?
		if (LESS (y, ey0, ysquare, esquare)
		    && LESS (ex0, x, esquare, xsquare)
		    && LESS (x, ex1, xsquare, esquare))
		    RET     break;
	    case 3:		// right (vertical) edge.
		if (LESS (x, ex0, xsquare, esquare)
		    && LESS (ey0, y, esquare, ysquare)
		    && LESS (y, ey1, ysquare, esquare))
		    RET     break;
	    }
      }
}


#ifdef DPER
#define MODPER(dx,dy) PRINTIOCC(perimeter); PRINTIOCC(dx); PRINTIOC(dy); cout <<endl; perimeter += (dx)+(dy);
#else
#define MODPER(dx,dy) perimeter += (dx)+(dy);
#endif

// Process one cell, finding intersections, and updating total area
// and perimeter.

void
process_one_cell (const int icell, const int jcell)
{
    const   Cell & cell (grid[icell][jcell]);
    const int n (cell.n ());

    // Compare the edge segments in this cell, pair by pair.  Exactly
    // one edge must be vertical.  This would be faster if we first
    // partitioned the edges into a horizontal set and a vertical set,
    // then compared only horizontal edges with vertical ones.
    // However, would the difference be significant?

#ifdef DPC
    cout << "Processing cell (" << icell << ", " << jcell << "), with "
	<< n << " edges:\n";
#endif

    if (n <= 1)
	return;

#ifdef WATCHDOG
    if (icell%10==0 && jcell==0) {
    ostrstream sout;
    sout << "process_one_cell: " << icell << ' ' << jcell << ends;
    watchdog_print(sout.str());
    }
#endif

    for (int i = 0; i < n; i++)
      {
	  const int e = cell[i];
#ifdef DPC
	  cout << i << " ";
#endif
	  const short int cx0 = determine_1d_cell (edge_x0 (e));
	  const short int cy0 = determine_1d_cell (edge_y0 (e));
	  const short int cx1 = determine_1d_cell (edge_x1 (e));
	  const short int cy1 = determine_1d_cell (edge_y1 (e));
#ifdef DPC
	  cout << e << ": (" << edge_x0 (e) << ',' << edge_y0 (e) << "), (" <<
	      edge_x1 (e) << ',' << edge_y1 (e) << ')';
	  cout << " = (" << cx0 << ',' << cy0 << "), (" << cx1 << ',' << cy1 << ')';
	  cout << ' ' << (edge_x0 (e) == edge_x1 (e) ? 'v' : 'h') << endl;
#endif
      }

    for (int i = 0; i < n; i++)
      {
	  const int ehor = cell[i];
	  if (ehor % 2)
	    {
#ifdef DPC
		cout << "Edge " << i << ' ' << ehor << " is not hor."
		    << (ehor % 2) << endl;
#endif
		continue;
	    }

	  const short int y = edge_y0 (ehor);

	  for (int j = 0; j < n; j++)
	    {			// Now comparing edges i and j of this cell.
		const int evert = cell[j];
		if (!(evert % 2))
		    continue;
		// ehor is horizontal; evert vertical.
#ifdef DPC
		cout << "Edges " << i << ' ' << j << endl;
#endif
		if (ehor / 4 == evert / 4)
		  {
#ifdef DPC
		      cout << "Edges " << i << " and " << j << " are from square #"
			  << (ehor / 4) << endl;
#endif
		      continue;	// Edges from the same square don't intersect.
		  }

		const short int x = edge_x0 (evert);

		if (y < edge_y0 (evert))
		    continue;
		if (y == edge_y0 (evert) && ehor < evert)
		    continue;
		if (y > edge_y1 (evert))
		    continue;
		if (y == edge_y1 (evert) && ehor > evert)
		    continue;
		if (x < edge_x0 (ehor))
		    continue;
		if (x == edge_x0 (ehor) && ehor > evert)
		    continue;
		if (x > edge_x1 (ehor))
		    continue;
		if (x == edge_x1 (ehor) && ehor < evert)
		    continue;

		nxsectincdup++;

		// We have an intersection, at (x,y); is it in this 
		// cell? In the special case of isothetic edges, it will always be.

#ifdef DPC
		cout << "Edges " << i << " and " << j << " intersect " << endl;
#endif

#ifdef NOTISOTHETIC
		const int cx = determine_1d_cell (x);
		const int cy = determine_1d_cell (y);
#ifdef DPC
		cout << "in cell (" << cx << ',' << cy << ")." << endl;
#endif
		if (cx != icell || cy != jcell)
		  {
#ifdef DPC
		      cout << "However, we're in cell (" << icell << ", " <<
			  jcell << ')' << endl;
#endif
		      continue;
		  }
#endif
		nxsect++;
		bool    visp = visible (x, y, evert / 4, ehor / 4);
#ifdef DPC
		cout << "visible? " << visp << endl;
#endif
		if (visp)
		  {		// This intersection is visible.
		      nvisxsect++;
		      const long int xs = (evert % 4 <= 1) ? x : -x;
		      const long int ys = (ehor % 4 <= 1) ? y : -y;
		      area -= xs * ys;
		  MODPER (xs, ys)}
	    }
      }
}


// Process the vertices in one square, testing if they are outside all
// other squares, and then calculating their contribution to the total
// area and perimeter.

void
process_vertices_in_one_square (const int i)
{
    Square & square = squares[i];
    const long int x0 = square.x;
    const long int y0 = square.y;
    const long int x1 = square.x + edgelen;
    const long int y1 = square.y + edgelen;

#ifdef DV
    cout << "Processing verts in square " << i << endl;
#endif

#ifdef WATCHDOG
    if (i%100000==0) {
    ostrstream sout;
    sout << "process_vertices_in_one_square: " << i << ends;
    watchdog_print(sout.str());
    }
#endif

    if (visible (x0, y0, i, i))
      {
	  area += x0 * y0;
	  // perimeter += -x0 - y0;
      MODPER (-x0, -y0)}
    if (visible (x1, y0, i, i))
      {
	  area -= x1 * y0;
	  // perimeter += x1 - y0;
      MODPER (x1, -y0)}
    if (visible (x0, y1, i, i))
      {
	  area -= x0 * y1;
	  // perimeter += -x0 + y1;
      MODPER (-x0, y1)}
    if (visible (x1, y1, i, i))
      {
	  area += x1 * y1;
	  // perimeter += x1 + y1;
      MODPER (x1, y1)}
}



void
print_stats1 ()
{
    const double edgelenfrac =
	static_cast < double >(edgelen) / static_cast < double >(MAXCOORD);

    const double gridperedge =
	1.0 / (edgelenfrac * static_cast < double >(gridsize));

    const double covcellfrac =
	static_cast < double >(nuniquecovers) * 100.0 / static_cast <
	double >(gridsize * gridsize);

    const double covsegfrac =
	static_cast < double >(ncoverededgesegs) / static_cast <
	double >(nedgesegs) * 100.0;

    const double segsperedge =
	static_cast < double >(nedgesegs) / static_cast < double >(nedges);

    const double segspercell =
	static_cast < double >(nedgesegs) / static_cast <
	double >(gridsize * gridsize);

    const double reallocfrac =
	static_cast < double >(nreallocedgecopies) / static_cast <
	double >(nedgesegs) * 100.0;

    const double depth = edgelenfrac * edgelenfrac * nsquares;

    const long int predicted_area
	=
	static_cast < long int >(static_cast <
				 double >(MAXCOORD * MAXCOORD) * (1 -
								  pow (1 -
								       edgelenfrac *
								       edgelenfrac,
								       nsquares)));

    const double pred_xsect = pow (nedges * edgelenfrac / 2, 2);

    PRINTIOCC (nsquares);
    PRINTIOC (nedges);
    cout << endl;
    PRINTIOC (edgelen);
    cout << " (" << edgelenfrac << "=1/" << 1 / edgelenfrac << " of universe), \n";
    cout << "Depth= " << depth << endl;
    PRINTIOC (predicted_area);
    cout << endl;
    cout << "Predicted number of intersections (incl covered)=" << pred_xsect
	<< endl;
    PRINTIOC (gridsize);
    cout << " (" << gridperedge << " times edgelen), ";
    PRINTIOCC (ncovers);
    PRINTIOC (nuniquecovers);
    cout << " (" << covcellfrac << "% of grid)\n";
    PRINTIOC (nedgesegs);
    cout << " (" << segsperedge << " per edge, " << segspercell << " per cell), ";
    PRINTIOC (ncoverededgesegs);;
    cout << " (" << covsegfrac << "% of segs)\n";
    PRINTIOCC (nreallocs);
    PRINTIOC (nreallocedgecopies);
    cout << " (" << reallocfrac << "% of segs)" << endl;
}


void compute_cell_hist() {
  const int mcount=10;
  long int counts[mcount];
  for (int i=0; i<mcount; i++) counts[i] = 0;
  for (int icell=0; icell < gridsize; icell++)
    for (int jcell=0; jcell < gridsize; jcell++) {
      int n = max(0, min(mcount-1, grid[icell][jcell].n()));
      counts[n]++;
    }
  cout << "Number of cells with k edge segments\n";
  for (int i=0; i<mcount-1; i++)
    cout << i << ':' << counts[i] << ", ";
  cout << '>' << mcount-2 << ':' << counts[mcount-1] << endl;
}



void
print_stats2 ()
{
    PRINTIOCC (nxsect);
    PRINTIOCC (nxsectincdup);
    PRINTIOC (nvisxsect);
    cout << endl;
    PRINTIOCC (area);
    PRINTIOC (perimeter);
    cout << endl;
}

int
main (const int argc, const char **const argv)
{

    initialize (argc, argv);
    ptime ("initialization");
    setup_grid ();
    ptime ("setup_grid");
    make_random_squares ();
    ptime ("make_random_square");
    for (int i = 0; i < nsquares; i++)
      {
	  put_square_and_edges_into_cells (i);
      }
    ptime ("put_square_and_edges_into_cells");
    print_stats1 ();
    compute_cell_hist();
    for (int i = 0; i < nsquares; i++)
      {
	  process_vertices_in_one_square (i);
      }
    ptime ("process_vertices_in_one_square");
    for (int i = 0; i < gridsize; i++)
	for (int j = 0; j < gridsize; j++)
	  {
	      process_one_cell (i, j);
	  }
    ptime ("process_one_cell");
    print_stats2 ();
    ptime ("end");
}
